use core::str;
use std::{
  ffi::CStr,
  fmt::{Debug, Display},
};

use anyhow::{bail, Context};
use newtype_enum::newtype_enum;
use zerocopy::{
  FromBytes, FromZeros, Immutable, IntoBytes, KnownLayout, Unaligned,
};

#[derive(Debug, Clone, Copy, FromZeros, IntoBytes, KnownLayout, Immutable)]
#[repr(C, packed)]
pub struct PacketHeader {
  pub user_id: u128,
  pub kind: PacketKind,
  pub size: u16,
}

#[derive(Debug, Clone, Copy, FromZeros, IntoBytes, KnownLayout, Immutable)]
#[repr(u16)]
pub enum PacketKind {
  Unknown = 0,
  Init = 1,
  Player = 2,
  Cap = 3,
  Game = 4,
  Tag = 5,
  Connect = 6,
  Disconnect = 7,
  Costume = 8,
  Shine = 9,
  Capture = 10,
  ChangeStage = 11,
  Command = 12,
  UdpInit = 13,
  HolePunch = 14
}

const COSTUME_NAME_SIZE: usize = 0x20;
const CAP_ANIM_SIZE: usize = 0x30;
const STAGE_GAME_NAME_SIZE: usize = 0x40;
const STAGE_CHANGE_NAME_SIZE: usize = 0x30;
const STAGE_ID_SIZE: usize = 0x10;
const CLIENT_NAME_SIZE: usize = COSTUME_NAME_SIZE;

#[derive(Debug)]
pub struct Packet {
  pub user_id: u128,
  pub data: PacketData,
}

#[newtype_enum]
#[derive(Debug)]
pub enum PacketData {
  Unknown(Vec<u8>),
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Init {
    pub max_players: u16,
  },
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Player {
    position: [f32; 3],
    rotation: [f32; 4],
    weights: [f32; 6],
    action: u16,
    subaction: u16,
  },
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Cap {
    position: [f32; 3],
    rotation: [f32; 4],
    out: Bool,
    anim: String<CAP_ANIM_SIZE>,
  },
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Game {
    is_2d: u8,
    scenario_num: u8,
    stage: String<STAGE_GAME_NAME_SIZE>,
  },
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Tag {
    update_type: u8,
    is_it: Bool,
    seconds: u8,
    minutes: u16,
  },
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Connect {
    kind: ConnectionKind,
    max_player: u16,
    client_name: String<CLIENT_NAME_SIZE>,
  },
  Disconnect,
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Costume {
    body_name: String<COSTUME_NAME_SIZE>,
    cap_name: String<COSTUME_NAME_SIZE>,
  },
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Shine {
    shine_id: i32,
    is_grand: Bool,
  },
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  Capture {
    model: String<COSTUME_NAME_SIZE>,
  },
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  ChangeStage {
    stage: String<STAGE_CHANGE_NAME_SIZE>,
    id: String<STAGE_ID_SIZE>,
    scenario: i8,
    sub_scenario: u8,
  },
  Command,
  #[derive(FromZeros, IntoBytes, KnownLayout, Immutable)]
  #[repr(C, packed)]
  UdpInit {
    port: u16
  },
  HolePunch
}

pub enum TagUpdateBit {
  Time = 0,
  State = 1,
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, FromZeros, IntoBytes, KnownLayout, Immutable)]
#[repr(u32)]
pub enum ConnectionKind {
  New = 0,
  Old = 1,
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, FromBytes, IntoBytes, KnownLayout, Immutable, Unaligned)]
#[repr(C)]
pub struct Bool(u8);

impl Bool {
  pub fn new(value: bool) -> Bool {
    Bool(if value { 1 } else { 0 })
  }
  pub fn get(&self) -> bool {
    self.0 != 0
  }
}

impl From<bool> for Bool {
  fn from(value: bool) -> Self {
    Bool::new(value)
  }
}

impl From<Bool> for bool {
  fn from(value: Bool) -> Self {
    value.get()
  }
}

#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, FromBytes, IntoBytes, KnownLayout, Immutable, Unaligned)]
#[repr(C)]
pub struct String<const N: usize>([u8; N]);

impl<const N: usize> String<N> {
  pub fn to_str(&self) -> anyhow::Result<&str> {
    let cstr = CStr::from_bytes_until_nul(&self.0).context("interpreting bytes as c-string")?;
    cstr.to_str().context("verifying string has utf-8")
    // let str = str::from_utf8(&self.0).context("verifying string has utf-8")?;
    // Ok(str.trim_end_matches('\0'))
  }
  pub fn assert_valid(&self) -> anyhow::Result<()> {
    self.to_str().map(drop)
  }
}

impl<const N: usize> TryFrom<&str> for String<N> {
  type Error = anyhow::Error;
  fn try_from(value: &str) -> Result<Self, Self::Error> {
    let mut buf = [0; N];
    if value.len() > N {
      bail!("seggs")
    }
    
    value.write_to_prefix(&mut buf).unwrap();

    Ok(Self(buf))
  }
}

impl<const N: usize> Display for String<N> {
  fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
    Display::fmt(self.to_str().expect("failed to parse string"), f)
  }
}
impl<const N: usize> Debug for String<N> {
  fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
    Debug::fmt(self.to_str().expect("failed to parse string"), f)
  }
}
